"""
implements a constant propagation analysis
"""

from dataclasses import dataclass
from cfg import Label, LSkip, LSet, LTest, LErr
from iteration import Transfer
from syntax import *

@dataclass
class abstract_value[T]:
    """None indicates that we don't know the precise value"""
    value: T | None
    def __str__(self):
        return f"{self.value}"

type abstract_env = dict[str, abstract_value[int]]
"""mapping from variables to abstract values.

As with concrete environment, a variable not in the dictionary will
lead to an error if we try to obtain its value
"""
type state = abstract_env | None
"""abstract state is either an abstract env or bottom
"""

def eval_aexp(env: abstract_env, e: ArithExpr) -> abstract_value[int] | None:
    """evaluate an arithmetic expression in an abstract environment
    returns None in case of error
    """
    match e:
        case AECst(value): return abstract_value(value)
        case AEVar(var):
            if var in env: return env[var]
            else: return None
        case AEUop(uop,expr):
            res = eval_aexp(env,expr)
            if res is None: return None
            if res.value is None: return abstract_value(None)
            if uop == Uop.OPP: return abstract_value(-res.value)
            return None
        case AEBop(bop,left_expr,right_expr):
            v1 = eval_aexp(env,left_expr)
            v2 = eval_aexp(env,right_expr)
            if v1 is None or v2 is None: return None
            if v1.value is None or v2.value is None:
                return abstract_value(None)
            match bop:
                case Bop.ADD: return abstract_value(v1.value+v2.value)
                case Bop.MUL: return abstract_value(v1.value*v2.value)
                case Bop.DIV:
                    if v2.value == 0: return None
                    return abstract_value(v1.value//v2.value)
        case _: assert False

def eval_bexp(env: abstract_env, e: BoolExpr) -> abstract_value[bool] | None:
    """abstract evaluation of a boolean expression"""
    match e:
        case BEPlain(aexpr):
            res = eval_aexp(env, aexpr)
            if res is None: return None
            if res.value is None: return abstract_value(None)
            return abstract_value(res.value!=0)
        case BEEq(left_expr,right_expr):
            v1 = eval_aexp(env, left_expr)
            v2 = eval_aexp(env, right_expr)
            if v1 is None or v2 is None: return None
            if v1.value is None or v2.value is None: return abstract_value(None)
            return abstract_value(v1.value == v2.value)
        case BELeq(left_expr,right_expr):
            v1 = eval_aexp(env, left_expr)
            v2 = eval_aexp(env, right_expr)
            if v1 is None or v2 is None: return None
            if v1.value is None or v2.value is None: return abstract_value(None)
            return abstract_value(v1.value <= v2.value)
        case BENeg(expr):
            v = eval_bexp(env,expr)
            if v is None: return None
            if v.value is None: return v
            return abstract_value(not v.value)
        case _: assert False

class Constant_propagation(Transfer[state]):
    variables: frozenset[str]
    def __init__(self,instr):
        self.variables = variables_of_instr(instr)
    def bottom(self) -> state:
        return None
    def init_state(self) -> state:
        return {var: abstract_value(None) for var in self.variables}
    def join(self, s1: state, s2: state) -> state:
        if s1 is None:
            return s2
        if s2 is None:
            return s1
        result = {}
        for var in self.variables:
            v1 = s1.get(var, abstract_value(None))
            v2 = s2.get(var, abstract_value(None))
            if v1.value == v2.value:
                result[var] = v1
            else:
                result[var] = abstract_value(None)
        return result
    
    def included(self, s1: state, s2: state) -> bool:
        if s1 is None:
            return True
        if s2 is None:
            return False
        return all(
            s2.get(var, abstract_value(None)).value in {s1.get(var, abstract_value(None)).value, None}
            for var in self.variables
        )

    def tr_skip(self, s: state) -> state:
        return s

    def tr_set(self, s: state, v: str, e: ArithExpr) -> state:
        if s is None:
            return None
        new_env = s.copy()
        new_env[v] = eval_aexp(s, e) or abstract_value(None)
        return new_env

    def tr_test(self, s: state, c: BoolExpr) -> state:
        if s is None:
            return None

        res = eval_bexp(s, c)
        if res is None:
            return None
        if res.value is False:
            return None

        if isinstance(c, BEEq) and isinstance(c.left_expr, AEVar) and isinstance(c.right_expr, AECst):
            var = c.left_expr.var
            val = c.right_expr.value
            new_env = s.copy()
            new_env[var] = abstract_value(val)
            return new_env

        return s

    def tr_err(self, s: state, e: Expr) -> state:
        return None
